#ifndef GLASGOW_CONSTRAINT_SOLVER_GUARD_GCS_CONSTRAINTS_ALL_DIFFERENT_HH
#define GLASGOW_CONSTRAINT_SOLVER_GUARD_GCS_CONSTRAINTS_ALL_DIFFERENT_HH

#include <gcs/constraint.hh>
#include <gcs/constraints/all_different/gac_all_different.hh>
#include <gcs/variable_id.hh>
#include <vector>

namespace gcs
{
    using AllDifferent = GACAllDifferent;
}

#endif
