#ifndef GLASGOW_CONSTRAINT_SOLVER_GUARD_GCS_GCS_HH
#define GLASGOW_CONSTRAINT_SOLVER_GUARD_GCS_GCS_HH 1

#include <gcs/constraint.hh>
#include <gcs/current_state.hh>
#include <gcs/exception.hh>
#include <gcs/expression.hh>
#include <gcs/extensional.hh>
#include <gcs/integer.hh>
#include <gcs/problem.hh>
#include <gcs/proof.hh>
#include <gcs/search_heuristics.hh>
#include <gcs/solve.hh>
#include <gcs/stats.hh>
#include <gcs/variable_condition.hh>
#include <gcs/variable_id.hh>

#include <gcs/constraints/abs.hh>
#include <gcs/constraints/all_different.hh>
#include <gcs/constraints/among.hh>
#include <gcs/constraints/arithmetic.hh>
#include <gcs/constraints/circuit.hh>
#include <gcs/constraints/comparison.hh>
#include <gcs/constraints/count.hh>
#include <gcs/constraints/element.hh>
#include <gcs/constraints/equals.hh>
#include <gcs/constraints/inverse.hh>
#include <gcs/constraints/knapsack.hh>
#include <gcs/constraints/linear.hh>
#include <gcs/constraints/logical.hh>
#include <gcs/constraints/min_max.hh>
#include <gcs/constraints/minus.hh>
#include <gcs/constraints/mult_bc.hh>
#include <gcs/constraints/n_value.hh>
#include <gcs/constraints/not_equals.hh>
#include <gcs/constraints/parity.hh>
#include <gcs/constraints/plus.hh>
#include <gcs/constraints/regular.hh>
#include <gcs/constraints/smart_table.hh>
#include <gcs/constraints/table.hh>

#endif
